import math
import time

import pygame

from easypygamewidgets import font

pygame.init()

all_timekeepers = []


class Timekeeper:
    def __init__(self, screen: "easypygamewidgets.Screen | None" = None, auto_size: bool = True, width: int = 180,
                 height: int = 80, start_at: float | int = 60, end_at: float | int | None = None,
                 show_milliseconds: bool = False, show_seconds: bool = True,
                 show_minutes: bool = False, smart_minutes: bool = True, show_hours: bool = False,
                 smart_hours: bool = True,
                 state: str = "enabled",
                 active_unpressed_text_color: tuple = (255, 255, 255),
                 disabled_unpressed_text_color: tuple = (150, 150, 150),
                 active_hover_text_color: tuple = (255, 255, 255),
                 disabled_hover_text_color: tuple = (150, 150, 150),
                 active_pressed_text_color: tuple = (200, 200, 200),
                 active_unpressed_background_color: tuple = (50, 50, 50),
                 disabled_unpressed_background_color: tuple = (30, 30, 30),
                 active_hover_background_color: tuple = (50, 50, 50),
                 disabled_hover_background_color: tuple = (30, 30, 30),
                 active_pressed_background_color: tuple = (40, 40, 40),
                 active_unpressed_border_color: tuple = (100, 100, 100),
                 disabled_unpressed_border_color: tuple = (60, 60, 60),
                 active_hover_border_color: tuple = (100, 100, 100),
                 disabled_hover_border_color: tuple = (60, 60, 60),
                 active_pressed_border_color: tuple = (50, 50, 50),
                 border_thickness: int = 2,
                 active_hover_cursor: pygame.Cursor = None,
                 disabled_hover_cursor: pygame.Cursor = None,
                 active_pressed_cursor: pygame.Cursor = None,
                 font: pygame.font.Font = font.default_font, alignment: str = "center",
                 alignment_spacing: int = 20, corner_radius: int = 14, ticking: bool = False,
                 type_order: list[str] = ("h", ":", "m", ":", "s", ".", "ms"), reversed: bool = False):
        if screen:
            screen.add_widget(self)
            self.screen = screen
        else:
            self.screen = None
            self.visible = True
            self.state = state
        self.auto_size = auto_size
        self.width = width
        self.height = height
        self.start_at = start_at
        self.end_at = end_at
        self.show_milliseconds = show_milliseconds
        self.show_seconds = show_seconds
        self.show_minutes = show_minutes
        self.smart_minutes = smart_minutes
        self.show_hours = show_hours
        self.smart_hours = smart_hours
        self.active_unpressed_text_color = active_unpressed_text_color
        self.disabled_unpressed_text_color = disabled_unpressed_text_color
        self.active_hover_text_color = active_hover_text_color
        self.disabled_hover_text_color = disabled_hover_text_color
        self.active_pressed_text_color = active_pressed_text_color
        self.active_unpressed_background_color = active_unpressed_background_color
        self.disabled_unpressed_background_color = disabled_unpressed_background_color
        self.active_hover_background_color = active_hover_background_color
        self.disabled_hover_background_color = disabled_hover_background_color
        self.active_pressed_background_color = active_pressed_background_color
        self.active_unpressed_border_color = active_unpressed_border_color
        self.disabled_unpressed_border_color = disabled_unpressed_border_color
        self.active_hover_border_color = active_hover_border_color
        self.disabled_hover_border_color = disabled_hover_border_color
        self.active_pressed_border_color = active_pressed_border_color
        self.border_thickness = border_thickness
        cursor_input = {
            "active_hover": active_hover_cursor,
            "disabled_hover": disabled_hover_cursor,
            "active_pressed": active_pressed_cursor
        }
        self.cursors = {}
        for name, cursor in cursor_input.items():
            if isinstance(cursor, pygame.cursors.Cursor):
                self.cursors[name] = cursor
            else:
                if cursor is not None:
                    print(
                        f"No custom cursor is used for the timekeeper {self.text} because it's not a pygame.Cursor object. ({cursor})")
                self.cursors[name] = None
        self.font = font
        self.alignment = alignment
        self.alignment_spacing = alignment_spacing
        self.corner_radius = corner_radius
        self.ticking = ticking
        self.type_order = type_order
        self.reversed = reversed
        self.x = 0
        self.y = 0
        self.alive = True
        self.pressed = False
        self.rect = pygame.Rect(self.x, self.y, self.width, self.height)
        self.original_cursor = None
        self.last_updated = None
        self.is_negative = False
        self.bindings = {}

        split_to_values(self, start_at)

        all_timekeepers.append(self)

    def configure(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        if 'x' in kwargs or 'y' in kwargs or 'width' in kwargs or 'height' in kwargs:
            self.rect = pygame.Rect(self.x, self.y, self.width, self.height)
        if 'screen' in kwargs:
            self.set_screen(kwargs["screen"])
        return self

    def config(self, **kwargs):
        self.configure(**kwargs)

    def delete(self):
        self.alive = False
        if self in all_timekeepers:
            all_timekeepers.remove(self)

    def place(self, x: int, y: int):
        self.x = x
        self.y = y
        self.rect = pygame.Rect(self.x, self.y, self.width, self.height)
        return self

    def bind(self, event: str, command, require_hover: bool = True):
        self.bindings[event] = {"command": command, "require_hover": require_hover}
        return self

    def trigger_event(self, event: str, *args, **kwargs):
        if event in self.bindings:
            binding_data = self.bindings[event]
            command = binding_data["command"]
            require_hover = binding_data["require_hover"]
            if not require_hover or is_point_in_rounded_rect(self, pygame.mouse.get_pos()):
                command(*args, **kwargs)

    def get_display_text(self):
        values = {
            "ms": self.milliseconds if self.show_milliseconds else None,
            "s": self.seconds if self.show_seconds else None,
            "m": self.minutes if (self.show_minutes or (self.smart_minutes and self.minutes != 0)) else None,
            "h": self.hours if (self.show_hours or (self.smart_hours and self.hours != 0)) else None,
        }
        parts = []
        pending_sep = None
        for token in self.type_order:
            if token in values:
                value = values[token]
                if value is not None:
                    if pending_sep and parts:
                        parts.append(pending_sep)
                    if token == "ms":
                        ms_int = int(round(value, 4) * 100) % 100
                        parts.append(f"{ms_int:02}")
                    else:
                        parts.append(f"{value:02}")
                    pending_sep = None
            else:
                pending_sep = token
        display_str = "".join(parts)
        if self.is_negative:
            display_str = "-" + display_str
        return display_str

    def set(self, milliseconds=0, seconds=0, minutes=0, hours=0):
        split_to_values(self, hours * 3600 + minutes * 60 + seconds + milliseconds / 1000)
        return self

    def stop(self):
        self.ticking = False
        return self

    def resume(self):
        self.ticking = True
        return self

    def start(self):
        self.ticking = True
        return self

    def reset(self):
        split_to_values(self, self.start_at)
        return self

    def add(self, amount):
        sign = -1 if self.is_negative else 1
        curr = ((self.hours * 3600) + (self.minutes * 60) + self.seconds + self.milliseconds) * sign
        curr += amount
        split_to_values(self, curr)
        return self

    def subtract(self, amount):
        sign = -1 if self.is_negative else 1
        curr = ((self.hours * 3600) + (self.minutes * 60) + self.seconds + self.milliseconds) * sign
        curr -= amount
        split_to_values(self, curr)
        return self

    def set_screen(self, screen):
        if self.screen:
            if self in screen.widgets:
                self.screen.widgets.remove(self)
        self.screen = screen
        screen.add_widget(self)
        return self

    def unbind(self, event: str):
        if event in self.bindings:
            del self.bindings[event]
        return self

    def unbind_all(self):
        self.bindings.clear()
        return self


def get_screen_offset(widget):
    if widget.screen:
        return widget.screen.x, widget.screen.y
    return 0, 0


def split_to_values(widget, total_seconds):
    base_seconds = math.floor(total_seconds)
    widget.is_negative = base_seconds < 0
    abs_secs = abs(base_seconds)
    widget.hours = int(abs_secs // 3600)
    widget.minutes = int((abs_secs % 3600) // 60)
    widget.seconds = int(abs_secs % 60)
    widget.milliseconds = abs(total_seconds) - int(abs(abs_secs))


def draw(timekeeper, surface: pygame.Surface):
    if not timekeeper.alive or not timekeeper.visible:
        return
    offset_x, offset_y = get_screen_offset(timekeeper)
    mouse_pos = pygame.mouse.get_pos()
    is_hovering = is_point_in_rounded_rect(timekeeper, mouse_pos)
    if timekeeper.state == "enabled":
        if timekeeper.pressed and is_hovering:
            text_color = timekeeper.active_pressed_text_color
            bg_color = timekeeper.active_pressed_background_color
            brd_color = timekeeper.active_pressed_border_color
        elif is_hovering:
            text_color = timekeeper.active_hover_text_color
            bg_color = timekeeper.active_hover_background_color
            brd_color = timekeeper.active_hover_border_color
        else:
            text_color = timekeeper.active_unpressed_text_color
            bg_color = timekeeper.active_unpressed_background_color
            brd_color = timekeeper.active_unpressed_border_color
    else:
        if is_hovering:
            text_color = timekeeper.disabled_hover_text_color
            bg_color = timekeeper.disabled_hover_background_color
            brd_color = timekeeper.disabled_hover_border_color
        else:
            text_color = timekeeper.disabled_unpressed_text_color
            bg_color = timekeeper.disabled_unpressed_background_color
            brd_color = timekeeper.disabled_unpressed_border_color

    if is_hovering:
        if timekeeper.state == "enabled":
            if timekeeper.pressed:
                cursor_key = "active_pressed"
            else:
                cursor_key = "active_hover"
        else:
            cursor_key = "disabled_hover"
        target_cursor = timekeeper.cursors.get(cursor_key)
        if target_cursor:
            current_cursor = pygame.mouse.get_cursor()
            if current_cursor != target_cursor:
                if timekeeper.original_cursor is None:
                    timekeeper.original_cursor = current_cursor
                pygame.mouse.set_cursor(target_cursor)
    else:
        if timekeeper.original_cursor:
            pygame.mouse.set_cursor(timekeeper.original_cursor)
            timekeeper.original_cursor = None

    if is_hovering and not getattr(timekeeper, "is_hovered", False):
        timekeeper.is_hovered = True
        timekeeper.trigger_event("<MOUSE-IN>")
    elif is_hovering and getattr(timekeeper, "is_hovered", False):
        timekeeper.is_hovered = True
        timekeeper.trigger_event("<HOVER>")
    elif not is_hovering and getattr(timekeeper, "is_hovered", False):
        timekeeper.is_hovered = False
        timekeeper.trigger_event("<MOUSE-OUT>")

    display_text = timekeeper.get_display_text()
    if timekeeper.auto_size:
        text_w = timekeeper.font.size(display_text)[0]
        exact_w = text_w + (timekeeper.alignment_spacing * 2)
        timekeeper.rect.width = (exact_w + 39) // 40 * 40
        timekeeper.rect.height = (timekeeper.font.size(display_text)[1] + 39) // 40 * 40
    draw_rect = timekeeper.rect.move(offset_x, offset_y)
    pygame.draw.rect(surface, bg_color, draw_rect, border_radius=timekeeper.corner_radius)
    if timekeeper.border_thickness > 0:
        pygame.draw.rect(surface, brd_color, draw_rect, width=timekeeper.border_thickness,
                         border_radius=timekeeper.corner_radius)
    old_clip = surface.get_clip()
    clip_rect = draw_rect.inflate(-4, -4)
    surface.set_clip(clip_rect)
    y_pos = draw_rect.centery
    drawn_stretched = False
    if timekeeper.alignment == "stretched" and len(display_text) > 1 and not timekeeper.auto_size:
        total_char_width = sum(timekeeper.font.render(char, True, text_color).get_width() for char in display_text)
        available_width = draw_rect.width - (timekeeper.alignment_spacing * 2)
        if available_width > total_char_width:
            drawn_stretched = True
            spacing = (available_width - total_char_width) / (len(display_text) - 1)
            current_x = draw_rect.left + timekeeper.alignment_spacing
            for char in display_text:
                char_surf = timekeeper.font.render(char, True, text_color)
                surface.blit(char_surf, char_surf.get_rect(midleft=(current_x, y_pos)))
                current_x += char_surf.get_width() + spacing
    if not drawn_stretched:
        text_surf = timekeeper.font.render(display_text, True, text_color)
        text_rect = text_surf.get_rect()
        if timekeeper.alignment == "left":
            text_rect.midleft = (draw_rect.left + timekeeper.alignment_spacing, y_pos)
        elif timekeeper.alignment == "right":
            text_rect.midright = (draw_rect.right - timekeeper.alignment_spacing, y_pos)
        else:
            text_rect.center = draw_rect.center
        surface.blit(text_surf, text_rect)
        timekeeper.last_text_x = text_rect.x
        surface.set_clip(old_clip)


def is_point_in_rounded_rect(timekeeper, point):
    offset_x, offset_y = get_screen_offset(timekeeper)
    rect = timekeeper.rect.move(offset_x, offset_y)
    if not rect.collidepoint(point): return False
    r = timekeeper.corner_radius
    r = min(r, rect.width // 2, rect.height // 2)
    if r <= 0: return True
    x, y = point
    if (rect.left + r <= x <= rect.right - r) or (rect.top + r <= y <= rect.bottom - r):
        return True
    centers = [(rect.left + r, rect.top + r), (rect.right - r, rect.top + r),
               (rect.left + r, rect.bottom - r), (rect.right - r, rect.bottom - r)]
    for cx, cy in centers:
        if ((x - cx) ** 2 + (y - cy) ** 2) <= r ** 2: return True
    return False


def react(timekeeper, event=None):
    if timekeeper.state != "enabled" or not timekeeper.visible:
        return
    is_inside = is_point_in_rounded_rect(timekeeper, pygame.mouse.get_pos())
    if event:
        if event.type == pygame.MOUSEBUTTONDOWN and event.button == 1 and is_inside:
            timekeeper.trigger_event("<PRESS>")
        elif event.type == pygame.MOUSEBUTTONUP and event.button == 1 and is_inside:
            timekeeper.trigger_event("<RELEASE>")
        elif event.type == pygame.KEYDOWN:
            timekeeper.trigger_event("<KEY>")
            if event.unicode:
                timekeeper.trigger_event(event.unicode)
            keyname = pygame.key.name(event.key)
            timekeeper.trigger_event(f"<{keyname.upper()}>")
    else:
        if timekeeper.ticking:
            now = time.time()
            if not timekeeper.last_updated:
                timekeeper.last_updated = now
            dt = now - timekeeper.last_updated
            timekeeper.last_updated = now
            sign = -1 if timekeeper.is_negative else 1
            curr = ((timekeeper.hours * 3600) + (
                    timekeeper.minutes * 60) + timekeeper.seconds + timekeeper.milliseconds) * sign
            change = -dt if timekeeper.reversed else dt
            next_value = curr + change
            if timekeeper.end_at is not None:
                reached_limit = False
                if not timekeeper.reversed and next_value >= timekeeper.end_at and not timekeeper.start_at > timekeeper.end_at:
                    reached_limit = True
                elif timekeeper.reversed and next_value <= timekeeper.end_at and not timekeeper.start_at < timekeeper.end_at:
                    reached_limit = True
                if reached_limit:
                    split_to_values(timekeeper, timekeeper.end_at)
                    timekeeper.stop()
                    timekeeper.trigger_event("<FINISHED>")
                    return
            split_to_values(timekeeper, next_value)



import time

import pygame

from easypygamewidgets import font

pygame.init()

all_labels = []


class Label:
    def __init__(self, screen: "easypygamewidgets.Screen | None" = None, auto_size: bool = True, width: int = 180,
                 height: int = 80,
                 text: str = "easypygamewidgets Label", state="enabled",
                 active_hover_text_color: tuple = (255, 255, 255),
                 active_hover_text_color_alpha: int = 255,
                 active_hover_shadow_color: tuple = (50, 50, 50),
                 active_hover_shadow_color_alpha: int = 200,
                 active_hover_background_color: tuple | None = None,
                 active_hover_background_color_alpha: int = 255,
                 active_hover_underline_color: tuple | None = None,
                 active_hover_underline_color_alpha: int = 255,
                 active_hover_strikethrough_color: tuple | None = None,
                 active_hover_strikethrough_color_alpha: int = 255,
                 active_hover_border_color: tuple = (),
                 active_hover_border_color_alpha: int = 255,
                 active_pressed_text_color: tuple = (255, 255, 255),
                 active_pressed_text_color_alpha: int = 255,
                 active_pressed_shadow_color: tuple = (50, 50, 50),
                 active_pressed_shadow_color_alpha: int = 200,
                 active_pressed_background_color: tuple | None = None,
                 active_pressed_background_color_alpha: int = 255,
                 active_pressed_underline_color: tuple | None = None,
                 active_pressed_underline_color_alpha: int = 255,
                 active_pressed_strikethrough_color: tuple | None = None,
                 active_pressed_strikethrough_color_alpha: int = 255,
                 active_pressed_border_color: tuple = (),
                 active_pressed_border_color_alpha: int = 255,
                 active_unpressed_text_color: tuple = (255, 255, 255),
                 active_unpressed_text_color_alpha: int = 255,
                 active_unpressed_shadow_color: tuple = (50, 50, 50),
                 active_unpressed_shadow_color_alpha: int = 200,
                 active_unpressed_background_color: tuple | None = None,
                 active_unpressed_background_color_alpha: int = 255,
                 active_unpressed_underline_color: tuple | None = None,
                 active_unpressed_underline_color_alpha: int = 255,
                 active_unpressed_strikethrough_color: tuple | None = None,
                 active_unpressed_strikethrough_color_alpha: int = 255,
                 active_unpressed_border_color: tuple = (),
                 active_unpressed_border_color_alpha: int = 255,
                 disabled_hover_text_color: tuple = (150, 150, 150),
                 disabled_hover_text_color_alpha: int = 255,
                 disabled_hover_shadow_color: tuple = (50, 50, 50),
                 disabled_hover_shadow_color_alpha: int = 200,
                 disabled_hover_background_color: tuple | None = None,
                 disabled_hover_background_color_alpha: int = 255,
                 disabled_hover_underline_color: tuple | None = None,
                 disabled_hover_underline_color_alpha: int = 255,
                 disabled_hover_strikethrough_color: tuple | None = None,
                 disabled_hover_strikethrough_color_alpha: int = 255,
                 disabled_hover_border_color: tuple = (),
                 disabled_hover_border_color_alpha: int = 255,
                 disabled_unpressed_text_color: tuple = (150, 150, 150),
                 disabled_unpressed_text_color_alpha: int = 255,
                 disabled_unpressed_shadow_color: tuple = (50, 50, 50),
                 disabled_unpressed_shadow_color_alpha: int = 200,
                 disabled_unpressed_background_color: tuple | None = None,
                 disabled_unpressed_background_color_alpha: int = 255,
                 disabled_unpressed_underline_color: tuple | None = None,
                 disabled_unpressed_underline_color_alpha: int = 255,
                 disabled_unpressed_strikethrough_color: tuple | None = None,
                 disabled_unpressed_strikethrough_color_alpha: int = 255,
                 disabled_unpressed_border_color: tuple = (),
                 disabled_unpressed_border_color_alpha: int = 255,
                 border_thickness: int = 2,
                 active_hover_cursor: pygame.Cursor = None,
                 disabled_hover_cursor: pygame.Cursor = None,
                 active_pressed_cursor: pygame.Cursor = None,
                 font: pygame.font.Font = font.default_font, alignment: str = "center",
                 alignment_spacing: int = 20, dragable: bool = False, top_left_corner_radius: int = 25,
                 top_right_corner_radius: int = 25, bottom_left_corner_radius: int = 25,
                 bottom_right_corner_radius: int = 25):
        tmp = font.render(text, True, (255, 255, 255))
        if screen:
            screen.add_widget(self)
            self.screen = screen
        else:
            self.screen = None
            self.visible = True
            self.state = state
        self.strikethrough = False
        self.underline = False
        self.auto_size = auto_size
        if auto_size:
            self.width = tmp.get_width() + 40 + (alignment_spacing - 20)
            self.height = tmp.get_height() + 20
        else:
            self.width = width
            self.height = height
        self.text = text
        self.active_hover_text_color = active_hover_text_color
        self.active_hover_text_color_alpha = active_hover_text_color_alpha
        self.active_hover_shadow_color = active_hover_shadow_color
        self.active_hover_shadow_color_alpha = active_hover_shadow_color_alpha
        self.active_hover_background_color = active_hover_background_color
        self.active_hover_background_color_alpha = active_hover_background_color_alpha
        if active_hover_underline_color:
            self.active_hover_underline_color = active_hover_underline_color
            self.underline = True
        else:
            self.active_hover_underline_color = active_hover_text_color
        self.active_hover_underline_color_alpha = active_hover_underline_color_alpha
        if active_hover_strikethrough_color:
            self.active_hover_strikethrough_color = active_hover_strikethrough_color
            self.strikethrough = True
        else:
            self.active_hover_strikethrough_color = active_hover_text_color
        self.active_hover_strikethrough_color_alpha = active_hover_strikethrough_color_alpha
        self.active_hover_border_color = active_hover_border_color
        self.active_hover_border_color_alpha = active_hover_border_color_alpha
        self.active_pressed_text_color = active_pressed_text_color
        self.active_pressed_text_color_alpha = active_pressed_text_color_alpha
        self.active_pressed_shadow_color = active_pressed_shadow_color
        self.active_pressed_shadow_color_alpha = active_pressed_shadow_color_alpha
        self.active_pressed_background_color = active_pressed_background_color
        self.active_pressed_background_color_alpha = active_pressed_background_color_alpha
        if active_pressed_underline_color:
            self.active_pressed_underline_color = active_pressed_underline_color
            self.underline = True
        else:
            self.active_pressed_underline_color = active_pressed_text_color
        self.active_pressed_underline_color_alpha = active_pressed_underline_color_alpha
        if active_pressed_strikethrough_color:
            self.active_pressed_strikethrough_color = active_pressed_strikethrough_color
            self.strikethrough = True
        else:
            self.active_pressed_strikethrough_color = active_pressed_text_color
        self.active_pressed_strikethrough_color_alpha = active_pressed_strikethrough_color_alpha
        self.active_pressed_border_color = active_pressed_border_color
        self.active_pressed_border_color_alpha = active_pressed_border_color_alpha
        self.active_unpressed_text_color = active_unpressed_text_color
        self.active_unpressed_text_color_alpha = active_unpressed_text_color_alpha
        self.active_unpressed_shadow_color = active_unpressed_shadow_color
        self.active_unpressed_shadow_color_alpha = active_unpressed_shadow_color_alpha
        self.active_unpressed_background_color = active_unpressed_background_color
        self.active_unpressed_background_color_alpha = active_unpressed_background_color_alpha
        if active_unpressed_underline_color:
            self.active_unpressed_underline_color = active_unpressed_underline_color
            self.underline = True
        else:
            self.active_unpressed_underline_color = active_unpressed_text_color
        self.active_unpressed_underline_color_alpha = active_unpressed_underline_color_alpha
        if active_unpressed_strikethrough_color:
            self.active_unpressed_strikethrough_color = active_unpressed_strikethrough_color
            self.strikethrough = True
        else:
            self.active_unpressed_strikethrough_color = active_unpressed_text_color
        self.active_unpressed_strikethrough_color_alpha = active_unpressed_strikethrough_color_alpha
        self.active_unpressed_border_color = active_unpressed_border_color
        self.active_unpressed_border_color_alpha = active_unpressed_border_color_alpha
        self.disabled_hover_text_color = disabled_hover_text_color
        self.disabled_hover_text_color_alpha = disabled_hover_text_color_alpha
        self.disabled_hover_shadow_color = disabled_hover_shadow_color
        self.disabled_hover_shadow_color_alpha = disabled_hover_shadow_color_alpha
        self.disabled_hover_background_color = disabled_hover_background_color
        self.disabled_hover_background_color_alpha = disabled_hover_background_color_alpha
        if disabled_hover_underline_color:
            self.disabled_hover_underline_color = disabled_hover_underline_color
            self.underline = True
        else:
            self.disabled_hover_underline_color = disabled_hover_text_color
        self.disabled_hover_underline_color_alpha = disabled_hover_underline_color_alpha
        if disabled_hover_strikethrough_color:
            self.disabled_hover_strikethrough_color = disabled_hover_strikethrough_color
            self.strikethrough = True
        else:
            self.disabled_hover_strikethrough_color = disabled_hover_text_color
        self.disabled_hover_strikethrough_color_alpha = disabled_hover_strikethrough_color_alpha
        self.disabled_hover_border_color = disabled_hover_border_color
        self.disabled_hover_border_color_alpha = disabled_hover_border_color_alpha
        self.disabled_unpressed_text_color = disabled_unpressed_text_color
        self.disabled_unpressed_text_color_alpha = disabled_unpressed_text_color_alpha
        self.disabled_unpressed_shadow_color = disabled_unpressed_shadow_color
        self.disabled_unpressed_shadow_color_alpha = disabled_unpressed_shadow_color_alpha
        self.disabled_unpressed_background_color = disabled_unpressed_background_color
        self.disabled_unpressed_background_color_alpha = disabled_unpressed_background_color_alpha
        if disabled_unpressed_underline_color:
            self.disabled_unpressed_underline_color = disabled_unpressed_underline_color
            self.underline = True
        else:
            self.disabled_unpressed_underline_color = disabled_unpressed_text_color
        self.disabled_unpressed_underline_color_alpha = disabled_unpressed_underline_color_alpha
        if disabled_unpressed_strikethrough_color:
            self.disabled_unpressed_strikethrough_color = disabled_unpressed_strikethrough_color
            self.strikethrough = True
        else:
            self.disabled_unpressed_strikethrough_color = disabled_unpressed_text_color
        self.disabled_unpressed_strikethrough_color_alpha = disabled_unpressed_strikethrough_color_alpha
        self.disabled_unpressed_border_color = disabled_unpressed_border_color
        self.disabled_unpressed_border_color_alpha = disabled_unpressed_border_color_alpha
        self.border_thickness = border_thickness
        cursor_input = {
            "active_hover": active_hover_cursor,
            "disabled_hover": disabled_hover_cursor,
            "active_pressed": active_pressed_cursor
        }
        self.cursors = {}
        for name, cursor in cursor_input.items():
            if isinstance(cursor, pygame.cursors.Cursor):
                self.cursors[name] = cursor
            else:
                if cursor is not None:
                    print(
                        f"No custom cursor is used for the label {self.text} because it's not a pygame.Cursor object. ({cursor})")
                self.cursors[name] = None
        self.font = font
        self.alignment = alignment
        self.alignment_spacing = alignment_spacing
        self.dragable = dragable
        self.top_left_corner_radius = top_left_corner_radius
        self.top_right_corner_radius = top_right_corner_radius
        self.bottom_left_corner_radius = bottom_left_corner_radius
        self.bottom_right_corner_radius = bottom_right_corner_radius
        self.x = 0
        self.y = 0
        self.alive = True
        self.pressed = False
        self.rect = pygame.Rect(self.x, self.y, self.width, self.height)
        self.original_cursor = None
        self.drag_offset = None
        self.is_dragging = False
        self.last_checked_dragging = None
        self.bindings = {}

        all_labels.append(self)

    def configure(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        if 'x' in kwargs or 'y' in kwargs or 'width' in kwargs or 'height' in kwargs or 'text' in kwargs:
            tmp = self.font.render(self.text, True, (255, 255, 255))
            self.rect = pygame.Rect(self.x, self.y, tmp.get_width(), tmp.get_height())
            self.width = self.rect.width + 40 + (self.alignment_spacing - 20)
            self.height = self.rect.height + 20
        if 'screen' in kwargs:
            self.set_screen(kwargs["screen"])
        return self

    def config(self, **kwargs):
        self.configure(**kwargs)

    def delete(self):
        self.alive = False
        if self in all_labels:
            all_labels.remove(self)

    def place(self, x: int, y: int):
        self.x = x
        self.y = y
        self.rect = pygame.Rect(self.x, self.y, self.width, self.height)
        return self

    def bind(self, event: str, command, require_hover: bool = True):
        self.bindings[event] = {"command": command, "require_hover": require_hover}
        return self

    def trigger_event(self, event: str, *args, **kwargs):
        if event in self.bindings:
            binding_data = self.bindings[event]
            command = binding_data["command"]
            require_hover = binding_data["require_hover"]
            if not require_hover or is_point_in_rounded_rect(self, pygame.mouse.get_pos()):
                command(*args, **kwargs)

    def set_screen(self, screen):
        if self.screen:
            if self in screen.widgets:
                self.screen.widgets.remove(self)
        self.screen = screen
        screen.add_widget(self)
        return self

    def set_strikethrough(self, value: bool):
        self.strikethrough = value
        return self

    def set_underline(self, value: bool):
        self.underline = value
        return self

    def unbind(self, event: str):
        if event in self.bindings:
            del self.bindings[event]
        return self

    def unbind_all(self):
        self.bindings.clear()
        return self


def get_screen_offset(widget):
    if widget.screen:
        return widget.screen.x, widget.screen.y
    return 0, 0


def combine_color_with_alpha(rgb, alpha):
    if not rgb: return None
    return (*rgb, alpha)


def draw(label, surface: pygame.Surface):
    if not label.alive or not label.visible:
        return
    mouse_pos = pygame.mouse.get_pos()
    is_hovering = is_point_in_rounded_rect(label, mouse_pos)
    if label.state == "enabled":
        if label.pressed:
            text_color = label.active_pressed_text_color
            bg_color = label.active_pressed_background_color
            shadow_color = label.active_pressed_shadow_color
            underline_color = label.active_pressed_underline_color
            strikethrough_color = label.active_pressed_strikethrough_color
            brd_color = label.active_pressed_border_color
            text_color_alpha = label.active_pressed_text_color_alpha
            bg_color_alpha = label.active_pressed_background_color_alpha
            shadow_color_alpha = label.active_pressed_shadow_color_alpha
            underline_color_alpha = label.active_pressed_underline_color_alpha
            strikethrough_color_alpha = label.active_pressed_strikethrough_color_alpha
            brd_color_alpha = label.active_pressed_border_color_alpha
        elif is_hovering:
            text_color = label.active_hover_text_color
            bg_color = label.active_hover_background_color
            shadow_color = label.active_hover_shadow_color
            underline_color = label.active_hover_underline_color
            strikethrough_color = label.active_hover_strikethrough_color
            brd_color = label.active_hover_border_color
            text_color_alpha = label.active_hover_text_color_alpha
            bg_color_alpha = label.active_hover_background_color_alpha
            shadow_color_alpha = label.active_hover_shadow_color_alpha
            underline_color_alpha = label.active_hover_underline_color_alpha
            strikethrough_color_alpha = label.active_hover_strikethrough_color_alpha
            brd_color_alpha = label.active_hover_border_color_alpha
        else:
            text_color = label.active_unpressed_text_color
            bg_color = label.active_unpressed_background_color
            shadow_color = label.active_unpressed_shadow_color
            underline_color = label.active_unpressed_underline_color
            strikethrough_color = label.active_unpressed_strikethrough_color
            brd_color = label.active_unpressed_border_color
            text_color_alpha = label.active_unpressed_text_color_alpha
            bg_color_alpha = label.active_unpressed_background_color_alpha
            shadow_color_alpha = label.active_unpressed_shadow_color_alpha
            underline_color_alpha = label.active_unpressed_underline_color_alpha
            strikethrough_color_alpha = label.active_unpressed_strikethrough_color_alpha
            brd_color_alpha = label.active_unpressed_border_color_alpha
    else:
        if is_hovering:
            text_color = label.disabled_hover_text_color
            bg_color = label.disabled_hover_background_color
            shadow_color = label.disabled_hover_shadow_color
            underline_color = label.disabled_hover_underline_color
            strikethrough_color = label.disabled_hover_strikethrough_color
            brd_color = label.disabled_hover_border_color
            text_color_alpha = label.disabled_hover_text_color_alpha
            bg_color_alpha = label.disabled_hover_background_color_alpha
            shadow_color_alpha = label.disabled_hover_shadow_color_alpha
            underline_color_alpha = label.disabled_hover_underline_color_alpha
            strikethrough_color_alpha = label.disabled_hover_strikethrough_color_alpha
            brd_color_alpha = label.disabled_hover_border_color_alpha
        else:
            text_color = label.disabled_unpressed_text_color
            bg_color = label.disabled_unpressed_background_color
            shadow_color = label.disabled_unpressed_shadow_color
            underline_color = label.disabled_unpressed_underline_color
            strikethrough_color = label.disabled_unpressed_strikethrough_color
            brd_color = label.disabled_unpressed_border_color
            text_color_alpha = label.disabled_unpressed_text_color_alpha
            bg_color_alpha = label.disabled_unpressed_background_color_alpha
            shadow_color_alpha = label.disabled_unpressed_shadow_color_alpha
            underline_color_alpha = label.disabled_unpressed_underline_color_alpha
            strikethrough_color_alpha = label.disabled_unpressed_strikethrough_color_alpha
            brd_color_alpha = label.disabled_unpressed_border_color_alpha

    text_color = combine_color_with_alpha(text_color, text_color_alpha)
    bg_color = combine_color_with_alpha(bg_color, bg_color_alpha)
    shadow_color = combine_color_with_alpha(shadow_color, shadow_color_alpha)
    underline_color = combine_color_with_alpha(underline_color, underline_color_alpha)
    strikethrough_color = combine_color_with_alpha(strikethrough_color, strikethrough_color_alpha)
    brd_color = combine_color_with_alpha(brd_color, brd_color_alpha)

    if is_hovering:
        if label.state == "enabled":
            if label.pressed:
                cursor_key = "active_pressed"
            else:
                cursor_key = "active_hover"
        else:
            cursor_key = "disabled_hover"
        target_cursor = label.cursors.get(cursor_key)
        if target_cursor:
            current_cursor = pygame.mouse.get_cursor()
            if current_cursor != target_cursor:
                if label.original_cursor is None:
                    label.original_cursor = current_cursor
                pygame.mouse.set_cursor(target_cursor)
    else:
        if label.original_cursor:
            pygame.mouse.set_cursor(label.original_cursor)
            label.original_cursor = None

    if is_hovering and not getattr(label, "is_hovered", False):
        label.is_hovered = True
        label.trigger_event("<MOUSE-IN>")
    elif is_hovering and getattr(label, "is_hovered", False):
        label.is_hovered = True
        label.trigger_event("<HOVER>")
    elif not is_hovering and getattr(label, "is_hovered", False):
        label.is_hovered = False
        label.trigger_event("<MOUSE-OUT>")

    if label.auto_size:
        temp_surf = label.font.render(label.text, True, text_color)
        label.width = temp_surf.get_width() + 40 + (label.alignment_spacing - 20)
        label.height = temp_surf.get_height() + 20
        label.rect = pygame.Rect(label.x, label.y, label.width, label.height)

    offset_x, offset_y = get_screen_offset(label)
    draw_rect = label.rect.move(offset_x, offset_y)

    draw_req_rect = pygame.Rect(0, 0, draw_rect.width, draw_rect.height)
    if bg_color:
        shape_surf = pygame.Surface((draw_rect.width, draw_rect.height), pygame.SRCALPHA)
        pygame.draw.rect(shape_surf, bg_color, draw_req_rect,
                         border_top_left_radius=label.top_left_corner_radius,
                         border_top_right_radius=label.top_right_corner_radius,
                         border_bottom_left_radius=label.bottom_left_corner_radius,
                         border_bottom_right_radius=label.bottom_right_corner_radius)
        shape_surf.set_alpha(bg_color[3])
        surface.blit(shape_surf, draw_rect)
    if brd_color:
        shape_surf = pygame.Surface((draw_rect.width, draw_rect.height), pygame.SRCALPHA)
        pygame.draw.rect(shape_surf, brd_color, draw_req_rect, width=label.border_thickness,
                         border_top_left_radius=label.top_left_corner_radius,
                         border_top_right_radius=label.top_right_corner_radius,
                         border_bottom_left_radius=label.bottom_left_corner_radius,
                         border_bottom_right_radius=label.bottom_right_corner_radius)
        shape_surf.set_alpha(bg_color[3])
        surface.blit(shape_surf, draw_rect)

    def render_text_line(txt, color, rect_ref, offset=(0, 0)):
        cx, cy = rect_ref.centerx + offset[0], rect_ref.centery + offset[1]
        if label.alignment == "stretched" and len(txt) > 1 and not label.auto_size:
            total_char_width = sum(label.font.render(char, True, color).get_width() for char in txt)
            available_width = rect_ref.width - (label.alignment_spacing * 2)
            if available_width > total_char_width:
                spacing = (available_width - total_char_width) / (len(txt) - 1)
                current_x = rect_ref.left + label.alignment_spacing + offset[0]
                for char in txt:
                    char_surf = label.font.render(char, True, color)
                    char_surf.set_alpha(color[3])
                    char_rect = char_surf.get_rect(midleft=(current_x, cy))
                    surface.blit(char_surf, char_rect)
                    current_x += char_surf.get_width() + spacing
                return pygame.Rect(rect_ref.left + label.alignment_spacing + offset[0],
                                   rect_ref.top + offset[1], available_width, rect_ref.height)
            txt_surf = label.font.render(txt, True, color)
            txt_rect = txt_surf.get_rect(center=(cx, cy))
        else:
            txt_surf = label.font.render(txt, True, color)
            txt_rect = txt_surf.get_rect()
            if label.alignment == "left":
                txt_rect.midleft = (rect_ref.left + label.alignment_spacing + offset[0], cy)
            elif label.alignment == "right":
                txt_rect.midright = (rect_ref.right - label.alignment_spacing + offset[0], cy)
            else:
                txt_rect.center = (cx, cy)
        txt_surf.set_alpha(color[3])
        surface.blit(txt_surf, txt_rect)
        return txt_rect

    if shadow_color and shadow_color[3] > 0:
        render_text_line(label.text, shadow_color, draw_rect, offset=(2, 2))
    final_text_rect = render_text_line(label.text, text_color, draw_rect)
    if final_text_rect:
        if underline_color and label.underline:
            shape_surf = pygame.Surface(final_text_rect.size, pygame.SRCALPHA)
            shape_surf_rect = shape_surf.get_rect()
            start_pos = (shape_surf_rect.left, shape_surf_rect.bottom - 2)
            end_pos = (shape_surf_rect.right, shape_surf_rect.bottom - 2)
            shape_surf.set_alpha(underline_color[3])
            pygame.draw.line(shape_surf, underline_color, start_pos, end_pos, 2)
            surface.blit(shape_surf, final_text_rect)
        if strikethrough_color and label.strikethrough:
            shape_surf = pygame.Surface(final_text_rect.size, pygame.SRCALPHA)
            shape_surf_rect = shape_surf.get_rect()
            start_pos = (shape_surf_rect.left, shape_surf_rect.centery)
            end_pos = (shape_surf_rect.right, shape_surf_rect.centery)
            shape_surf.set_alpha(strikethrough_color[3])
            pygame.draw.line(shape_surf, strikethrough_color, start_pos, end_pos, 2)
            surface.blit(shape_surf, final_text_rect)


def is_point_in_rounded_rect(label, point):
    offset_x, offset_y = get_screen_offset(label)
    rect = label.rect.move(offset_x, offset_y)
    if not rect.collidepoint(point): return False
    max_r = max(label.top_left_corner_radius, label.top_right_corner_radius,
                label.bottom_left_corner_radius, label.bottom_right_corner_radius)
    if (rect.left + max_r <= point[0] <= rect.right - max_r) or \
            (rect.top + max_r <= point[1] <= rect.bottom - max_r):
        return True
    x, y = point
    if x < rect.left + label.top_left_corner_radius and y < rect.top + label.top_left_corner_radius:
        cx, cy = rect.left + label.top_left_corner_radius, rect.top + label.top_left_corner_radius
        return (x - cx) ** 2 + (y - cy) ** 2 <= label.top_left_corner_radius ** 2
    if x > rect.right - label.top_right_corner_radius and y < rect.top + label.top_right_corner_radius:
        cx, cy = rect.right - label.top_right_corner_radius, rect.top + label.top_right_corner_radius
        return (x - cx) ** 2 + (y - cy) ** 2 <= label.top_right_corner_radius ** 2
    if x < rect.left + label.bottom_left_corner_radius and y > rect.bottom - label.bottom_left_corner_radius:
        cx, cy = rect.left + label.bottom_left_corner_radius, rect.bottom - label.bottom_left_corner_radius
        return (x - cx) ** 2 + (y - cy) ** 2 <= label.bottom_left_corner_radius ** 2
    if x > rect.right - label.bottom_right_corner_radius and y > rect.bottom - label.bottom_right_corner_radius:
        cx, cy = rect.right - label.bottom_right_corner_radius, rect.bottom - label.bottom_right_corner_radius
        return (x - cx) ** 2 + (y - cy) ** 2 <= label.bottom_right_corner_radius ** 2
    return True


def react(label, event=None):
    if label.state != "enabled" or not label.visible:
        label.pressed = False
        return
    current_time = time.time()
    mouse_pos = pygame.mouse.get_pos()
    is_inside = is_point_in_rounded_rect(label, mouse_pos)
    screen_off_x, screen_off_y = get_screen_offset(label)
    if event:
        if event.type == pygame.KEYDOWN:
            label.trigger_event("<KEY>")
            if event.unicode:
                label.trigger_event(event.unicode)
            keyname = pygame.key.name(event.key)
            label.trigger_event(f"<{keyname.upper()}>")
        if event.type == pygame.MOUSEMOTION:
            if label.pressed and label.dragable:
                if is_inside or label.is_dragging:
                    label.is_dragging = True
                    label.last_checked_dragging = current_time
                    if label.drag_offset:
                        new_x = mouse_pos[0] - label.drag_offset[0] - screen_off_x
                        new_y = mouse_pos[1] - label.drag_offset[1] - screen_off_y
                        label.place(new_x, new_y)
        elif event.type == pygame.MOUSEBUTTONDOWN and is_inside:
            if event.button == 1:
                label.pressed = True
                label.drag_offset = (mouse_pos[0] - (label.x + screen_off_x), mouse_pos[1] - (label.y + screen_off_y))
                label.trigger_event("<PRESS>")
        elif event.type == pygame.MOUSEBUTTONUP and is_inside:
            if event.button == 1:
                label.pressed = False
                label.is_dragging = False
                label.trigger_event("<RELEASE>")
    if label.last_checked_dragging:
        if current_time - label.last_checked_dragging > 0.2:
            label.is_dragging = False
    if label.pressed and not label.is_dragging:
        label.trigger_event("<HOLD>")
    if label.pressed and label.is_dragging:
        label.trigger_event("<DRAG>")



import sys

import pygame

from easypygamewidgets import font

pygame.init()

all_entrys = []


class Entry:
    def __init__(self, screen: "easypygamewidgets.Screen | None" = None, auto_size: bool = True, width: int = 180,
                 height: int = 80, placeholder_text: str = "easypygamewidgets Entry",
                 text: str = "", char_limit: int | None = None,
                 show: str | None = None, state: str = "enabled",
                 active_unpressed_text_color: tuple = (255, 255, 255),
                 disabled_unpressed_text_color: tuple = (150, 150, 150),
                 active_hover_text_color: tuple = (255, 255, 255),
                 disabled_hover_text_color: tuple = (150, 150, 150),
                 active_pressed_text_color: tuple = (200, 200, 200),
                 active_unpressed_background_color: tuple = (50, 50, 50),
                 disabled_unpressed_background_color: tuple = (30, 30, 30),
                 active_hover_background_color: tuple = (70, 70, 70),
                 disabled_hover_background_color: tuple = (30, 30, 30),
                 active_pressed_background_color: tuple = (40, 40, 40),
                 active_unpressed_border_color: tuple = (100, 100, 100),
                 disabled_unpressed_border_color: tuple = (60, 60, 60),
                 active_hover_border_color: tuple = (150, 150, 150),
                 disabled_hover_border_color: tuple = (60, 60, 60),
                 active_pressed_border_color: tuple = (50, 50, 50),
                 selection_color: tuple = (0, 120, 215),
                 disabled_selection_color: tuple = (32, 106, 163),
                 border_thickness: int = 2,
                 active_hover_cursor: pygame.Cursor = None,
                 disabled_hover_cursor: pygame.Cursor = None,
                 active_pressed_cursor: pygame.Cursor = None,
                 blinking_cursor: str = "|",
                 font: pygame.font.Font = font.default_font, alignment: str = "left",
                 alignment_spacing: int = 20, corner_radius: int = 25, repeat_delay: int = 500,
                 repeat_interval: int = 50):
        if screen:
            screen.add_widget(self)
            self.screen = screen
        else:
            self.screen = None
            self.visible = True
            self.state = state
        self.auto_size = auto_size
        self.width = width
        self.height = height
        self.placeholder_text = placeholder_text
        self.text = text
        self.char_limit = char_limit
        self.show = show
        self.active_unpressed_text_color = active_unpressed_text_color
        self.disabled_unpressed_text_color = disabled_unpressed_text_color
        self.active_hover_text_color = active_hover_text_color
        self.disabled_hover_text_color = disabled_hover_text_color
        self.active_pressed_text_color = active_pressed_text_color
        self.active_unpressed_background_color = active_unpressed_background_color
        self.disabled_unpressed_background_color = disabled_unpressed_background_color
        self.active_hover_background_color = active_hover_background_color
        self.disabled_hover_background_color = disabled_hover_background_color
        self.active_pressed_background_color = active_pressed_background_color
        self.active_unpressed_border_color = active_unpressed_border_color
        self.disabled_unpressed_border_color = disabled_unpressed_border_color
        self.active_hover_border_color = active_hover_border_color
        self.disabled_hover_border_color = disabled_hover_border_color
        self.active_pressed_border_color = active_pressed_border_color
        self.selection_color = selection_color
        self.disabled_selection_color = disabled_selection_color
        self.border_thickness = border_thickness
        cursor_input = {
            "active_hover": active_hover_cursor,
            "disabled_hover": disabled_hover_cursor,
            "active_pressed": active_pressed_cursor
        }
        self.cursors = {}
        for name, cursor in cursor_input.items():
            if isinstance(cursor, pygame.cursors.Cursor):
                self.cursors[name] = cursor
            else:
                if cursor is not None:
                    print(
                        f"No custom cursor is used for the entry {self.text} because it's not a pygame.Cursor object. ({cursor})")
                self.cursors[name] = None
        self.blinking_cursor = blinking_cursor
        self.font = font
        self.alignment = alignment
        self.alignment_spacing = alignment_spacing
        self.corner_radius = corner_radius
        self.repeat_delay = repeat_delay
        self.repeat_interval = repeat_interval
        self.x = 0
        self.y = 0
        self.alive = True
        self.pressed = False
        self.rect = pygame.Rect(self.x, self.y, self.width, self.height)
        self.original_cursor = None
        self.selected_text = None
        self.focused = False
        if text:
            self.cursor_position = len(text)
        else:
            self.cursor_position = 0
        self.scroll_offset = 0
        self.drag_start = None
        self.selection_anchor = None
        self.last_text_x = self.rect.left
        self.held_key_info = None
        self.next_repeat_time = 0
        self.cursor_visible = True
        self.last_blink_time = pygame.time.get_ticks()
        self.bindings = {}

        all_entrys.append(self)

    def configure(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        if 'x' in kwargs or 'y' in kwargs or 'width' in kwargs or 'height' in kwargs:
            self.rect = pygame.Rect(self.x, self.y, self.width, self.height)
        if 'screen' in kwargs:
            self.set_screen(kwargs["screen"])
        return self

    def config(self, **kwargs):
        self.configure(**kwargs)

    def delete(self):
        self.alive = False
        if self in all_entrys:
            all_entrys.remove(self)

    def place(self, x: int, y: int):
        self.x = x
        self.y = y
        self.rect = pygame.Rect(self.x, self.y, self.width, self.height)
        return self

    def bind(self, event: str, command, require_hover: bool = True):
        self.bindings[event] = {"command": command, "require_hover": require_hover}
        return self

    def trigger_event(self, event: str, *args, **kwargs):
        if event in self.bindings:
            binding_data = self.bindings[event]
            command = binding_data["command"]
            require_hover = binding_data["require_hover"]
            if not require_hover or is_point_in_rounded_rect(self, pygame.mouse.get_pos()):
                command(*args, **kwargs)

    def get(self):
        return self.text

    def text_delete(self, position_start: int = 0, position_end: int | None = None):
        if position_end is None:
            position_end = len(self.text)
        position_start = max(0, min(position_start, len(self.text)))
        position_end = max(0, min(position_end, len(self.text)))
        if position_start < position_end:
            self.text = self.text[:position_start] + self.text[position_end:]
            if self.cursor_position > position_end:
                self.cursor_position -= (position_end - position_start)
            elif self.cursor_position > position_start:
                self.cursor_position = position_start
        self.reset_cursor_blink()

    def text_insert(self, text: str, position: int = None):
        if position is None:
            position = len(self.text)
        if self.char_limit is not None and len(self.text) + len(text) > self.char_limit:
            return
        self.text = self.text[:position] + text + self.text[position:]
        self.cursor_position += len(text)
        self.reset_cursor_blink()

    def text_select(self, position_start: int = 0, position_end: int | None = None):
        if position_end is None:
            position_end = len(self.text)
        self.selected_text = [min(position_start, position_end), max(position_start, position_end)]
        self.reset_cursor_blink()

    def text_copy(self):
        if self.selected_text and self.selected_text[0] != self.selected_text[1]:
            start, end = self.selected_text
            clipboard_text = self.text[start:end]
            pygame.scrap.put(pygame.SCRAP_TEXT, clipboard_text.encode('utf-8'))

    def text_cut(self):
        if self.selected_text and self.selected_text[0] != self.selected_text[1]:
            self.text_copy()
            self.text_delete(self.selected_text[0], self.selected_text[1])
            self.selected_text = None

    def text_paste(self):
        if not pygame.scrap.get_init():
            pygame.scrap.init()
        if self.selected_text:
            self.text_delete(self.selected_text[0], self.selected_text[1])
            self.selected_text = None
        clipboard = pygame.scrap.get(pygame.SCRAP_TEXT)
        if clipboard:
            try:
                paste_text = clipboard.decode('utf-8').split('\x00')[0]
                self.text_insert(paste_text, self.cursor_position)
            except Exception as e:
                print(f"Paste error: {e}")

    def reset_cursor_blink(self):
        self.cursor_visible = True
        self.last_blink_time = pygame.time.get_ticks()

    def get_display_text(self):
        if self.text:
            if self.show:
                return self.show * len(self.text)
            return self.text
        elif self.placeholder_text and not self.focused:
            return self.placeholder_text
        return ""

    def set_screen(self, screen):
        if self.screen:
            if self in screen.widgets:
                self.screen.widgets.remove(self)
        self.screen = screen
        screen.add_widget(self)
        return self

    def unbind(self, event: str):
        if event in self.bindings:
            del self.bindings[event]
        return self

    def unbind_all(self):
        self.bindings.clear()
        return self


def process_key_action(entry, key, unicode_char):
    is_linux = sys.platform.startswith("linux")
    mods = pygame.key.get_mods()
    ctrl = (mods & pygame.KMOD_CTRL) or (mods & pygame.KMOD_META)
    shift = mods & pygame.KMOD_SHIFT
    entry.reset_cursor_blink()
    if key in (pygame.K_LEFT, pygame.K_RIGHT, pygame.K_HOME, pygame.K_END):
        if shift and entry.selected_text is None:
            entry.selection_anchor = entry.cursor_position
        if key == pygame.K_LEFT:
            entry.cursor_position = max(0, entry.cursor_position - 1)
        elif key == pygame.K_RIGHT:
            entry.cursor_position = min(len(entry.text), entry.cursor_position + 1)
        elif key == pygame.K_HOME:
            entry.cursor_position = 0
        elif key == pygame.K_END:
            entry.cursor_position = len(entry.text)
        if shift:
            entry.text_select(entry.selection_anchor, entry.cursor_position)
        else:
            entry.selected_text = None
            entry.selection_anchor = None
        return
    if ctrl:
        if not is_linux or shift:
            if key == pygame.K_c:
                entry.text_copy()
            elif key == pygame.K_v:
                entry.text_paste()
                entry.trigger_event("<PASTE>")
            elif key == pygame.K_x:
                entry.text_cut()
                entry.trigger_event("<CUT>")
        if key == pygame.K_a:
            entry.selection_anchor = 0
            entry.cursor_position = len(entry.text)
            entry.text_select(0, len(entry.text))
        return
    if key == pygame.K_BACKSPACE:
        if entry.selected_text:
            entry.text_delete(*entry.selected_text)
            entry.selected_text = None
        elif entry.cursor_position > 0:
            entry.text_delete(entry.cursor_position - 1, entry.cursor_position)
        entry.trigger_event("<BACKSPACE>")
        return
    elif key == pygame.K_DELETE:
        if entry.selected_text:
            entry.text_delete(*entry.selected_text)
            entry.selected_text = None
        elif entry.cursor_position < len(entry.text):
            entry.text_delete(entry.cursor_position, entry.cursor_position + 1)
        entry.trigger_event("<DELETE>")
        return
    elif unicode_char.isprintable() and unicode_char != "":
        if entry.selected_text:
            entry.text_delete(*entry.selected_text)
            entry.selected_text = None
        entry.text_insert(unicode_char, entry.cursor_position)
        entry.trigger_event("<TYPING>")


def get_screen_offset(widget):
    if widget.screen:
        return widget.screen.x, widget.screen.y
    return 0, 0


def draw(entry, surface: pygame.Surface):
    if not entry.alive or not entry.visible:
        return
    offset_x, offset_y = get_screen_offset(entry)
    if entry.focused and entry.held_key_info:
        current_time = pygame.time.get_ticks()
        if current_time >= entry.next_repeat_time:
            key, unicode_char = entry.held_key_info
            process_key_action(entry, key, unicode_char)
            entry.next_repeat_time = current_time + entry.repeat_interval
    if not pygame.scrap.get_init():
        pygame.scrap.init()
    mouse_pos = pygame.mouse.get_pos()
    is_hovering = is_point_in_rounded_rect(entry, mouse_pos)
    now = pygame.time.get_ticks()
    display_text = entry.get_display_text()
    if entry.state == "enabled":
        if entry.pressed and is_hovering:
            text_color = entry.active_pressed_text_color
            bg_color = entry.active_pressed_background_color
            brd_color = entry.active_pressed_border_color
        elif is_hovering:
            text_color = entry.active_hover_text_color
            bg_color = entry.active_hover_background_color
            brd_color = entry.active_hover_border_color
        else:
            text_color = entry.active_unpressed_text_color
            bg_color = entry.active_unpressed_background_color
            brd_color = entry.active_unpressed_border_color
        selection_color = entry.selection_color
    else:
        if is_hovering:
            text_color = entry.disabled_hover_text_color
            bg_color = entry.disabled_hover_background_color
            brd_color = entry.disabled_hover_border_color
        else:
            text_color = entry.disabled_unpressed_text_color
            bg_color = entry.disabled_unpressed_background_color
            brd_color = entry.disabled_unpressed_border_color
        selection_color = entry.disabled_selection_color

    if is_hovering:
        if entry.state == "enabled":
            if entry.pressed:
                cursor_key = "active_pressed"
            else:
                cursor_key = "active_hover"
        else:
            cursor_key = "disabled_hover"
        target_cursor = entry.cursors.get(cursor_key)
        if target_cursor:
            current_cursor = pygame.mouse.get_cursor()
            if current_cursor != target_cursor:
                if entry.original_cursor is None:
                    entry.original_cursor = current_cursor
                pygame.mouse.set_cursor(target_cursor)
    else:
        if entry.original_cursor:
            pygame.mouse.set_cursor(entry.original_cursor)
            entry.original_cursor = None

    if is_hovering and not getattr(entry, "is_hovered", False):
        entry.is_hovered = True
        entry.trigger_event("<MOUSE-IN>")
    elif is_hovering and getattr(entry, "is_hovered", False):
        entry.is_hovered = True
        entry.trigger_event("<HOVER>")
    elif not is_hovering and getattr(entry, "is_hovered", False):
        entry.is_hovered = False
        entry.trigger_event("<MOUSE-OUT>")

    if entry.auto_size:
        text_w = entry.font.size(display_text)[0]
        required_width = max(entry.width, text_w + (entry.alignment_spacing * 2) + 10)
        if entry.rect.width != required_width:
            entry.rect.width = required_width
    draw_rect = entry.rect.move(offset_x, offset_y)
    pygame.draw.rect(surface, bg_color, draw_rect, border_radius=entry.corner_radius)
    if entry.border_thickness > 0:
        pygame.draw.rect(surface, brd_color, draw_rect, width=entry.border_thickness,
                         border_radius=entry.corner_radius)
    old_clip = surface.get_clip()
    clip_rect = draw_rect.inflate(-4, -4)
    surface.set_clip(clip_rect)
    y_pos = draw_rect.centery
    drawn_stretched = False
    if entry.alignment == "stretched" and len(display_text) > 1 and not entry.auto_size:
        total_char_width = sum(entry.font.render(char, True, text_color).get_width() for char in display_text)
        available_width = draw_rect.width - (entry.alignment_spacing * 2)
        if available_width > total_char_width:
            drawn_stretched = True
            spacing = (available_width - total_char_width) / (len(display_text) - 1)
            current_x = draw_rect.left + entry.alignment_spacing
            for char in display_text:
                char_surf = entry.font.render(char, True, text_color)
                surface.blit(char_surf, char_surf.get_rect(midleft=(current_x, y_pos)))
                current_x += char_surf.get_width() + spacing
    if not drawn_stretched:
        text_surf = entry.font.render(display_text, True, text_color)
        text_rect = text_surf.get_rect()
        if entry.alignment == "left":
            text_rect.midleft = (draw_rect.left + entry.alignment_spacing, y_pos)
        elif entry.alignment == "right":
            text_rect.midright = (draw_rect.right - entry.alignment_spacing, y_pos)
        else:
            text_rect.center = draw_rect.center
        cursor_x_rel = entry.font.size(display_text[:entry.cursor_position])[0]
        visible_left = draw_rect.left + entry.alignment_spacing
        visible_right = draw_rect.right - entry.alignment_spacing
        visible_width = visible_right - visible_left
        if text_rect.width > visible_width and not entry.auto_size:
            text_rect.midleft = (draw_rect.left + entry.alignment_spacing, y_pos)
            text_rect.x += entry.scroll_offset
            cursor_screen_x = text_rect.x + cursor_x_rel
            if cursor_screen_x > visible_right:
                entry.scroll_offset -= (cursor_screen_x - visible_right)
            elif cursor_screen_x < visible_left:
                entry.scroll_offset += (visible_left - cursor_screen_x)
            if text_rect.x > visible_left:
                entry.scroll_offset = 0
                text_rect.x = visible_left
        else:
            entry.scroll_offset = 0
        if entry.selected_text and entry.selected_text[0] != entry.selected_text[1]:
            start_idx = min(entry.selected_text)
            end_idx = max(entry.selected_text)
            sel_start_x = entry.font.size(display_text[:start_idx])[0]
            sel_end_x = entry.font.size(display_text[:end_idx])[0]
            highlight_rect = pygame.Rect(text_rect.x + sel_start_x, text_rect.top, sel_end_x - sel_start_x,
                                         text_rect.height)
            pygame.draw.rect(surface, selection_color, highlight_rect)
        surface.blit(text_surf, text_rect)
        has_selection = entry.selected_text and entry.selected_text[0] != entry.selected_text[1]
        if entry.focused and entry.state == "enabled" and not has_selection:
            if now - entry.last_blink_time > 500:
                entry.cursor_visible = not entry.cursor_visible
                entry.last_blink_time = now
            if entry.cursor_visible:
                line_x = text_rect.x + cursor_x_rel
                if visible_left <= line_x <= visible_right + 2:
                    cursor_surf = entry.font.render(entry.blinking_cursor, True, text_color)
                    cursor_rect = cursor_surf.get_rect(midleft=(line_x, text_rect.centery))
                    surface.blit(cursor_surf, cursor_rect)
        entry.last_text_x = text_rect.x
        surface.set_clip(old_clip)


def is_point_in_rounded_rect(entry, point):
    offset_x, offset_y = get_screen_offset(entry)
    rect = entry.rect.move(offset_x, offset_y)
    if not rect.collidepoint(point): return False
    r = entry.corner_radius
    r = min(r, rect.width // 2, rect.height // 2)
    if r <= 0: return True
    x, y = point
    if (rect.left + r <= x <= rect.right - r) or (rect.top + r <= y <= rect.bottom - r):
        return True
    centers = [(rect.left + r, rect.top + r), (rect.right - r, rect.top + r),
               (rect.left + r, rect.bottom - r), (rect.right - r, rect.bottom - r)]
    for cx, cy in centers:
        if ((x - cx) ** 2 + (y - cy) ** 2) <= r ** 2: return True
    return False


def react(entry, event=None):
    if entry.state != "enabled" or not entry.visible:
        return
    display_text = entry.get_display_text()
    is_inside = is_point_in_rounded_rect(entry, pygame.mouse.get_pos())

    def get_idx_at_mouse(mouse_x):
        curr_x = entry.last_text_x
        for i, char in enumerate(display_text):
            char_w = entry.font.size(char)[0]
            if mouse_x < curr_x + char_w / 2: return i
            curr_x += char_w
        return min(len(display_text), len(entry.text))

    if event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
        if is_inside:
            entry.trigger_event("<PRESS>")
        if not entry.focused:
            entry.trigger_event("<FOCUS-IN>")
        if is_inside:
            entry.pressed = True
            idx = get_idx_at_mouse(event.pos[0])
            # This somehow has to be redone because """return min(len(display_text), len(entry.text))""" doesn't work
            entry.cursor_position = min(len(entry.text), idx)
            entry.selection_anchor = idx
            entry.selected_text = None
            if not entry.focused:
                entry.focused = True
            entry.reset_cursor_blink()
        else:
            if entry.focused:
                entry.trigger_event("<FOCUS-OUT>")
            entry.focused = False
    elif event.type == pygame.MOUSEBUTTONUP and event.button == 1:
        entry.trigger_event("<RELEASE>")
        entry.pressed = False
        entry.selection_anchor = None
    elif event.type == pygame.MOUSEMOTION and entry.pressed:
        if entry.selection_anchor is not None:
            entry.cursor_position = get_idx_at_mouse(event.pos[0])
            entry.text_select(entry.selection_anchor, entry.cursor_position)
            entry.reset_cursor_blink()
    elif event.type == pygame.KEYDOWN:
        if entry.focused:
            process_key_action(entry, event.key, event.unicode)
            entry.held_key_info = (event.key, event.unicode)
            entry.next_repeat_time = pygame.time.get_ticks() + entry.repeat_delay
        entry.trigger_event("<KEY>")
        if event.unicode:
            entry.trigger_event(event.unicode)
        keyname = pygame.key.name(event.key)
        entry.trigger_event(f"<{keyname.upper()}>")
    elif event.type == pygame.KEYUP:
        if entry.held_key_info and event.key == entry.held_key_info[0]:
            entry.held_key_info = None



import os
import pathlib

import pygame

pygame.init()
pack_font_path = pathlib.Path(__file__).resolve().parent / "assets" / "fonts"
default_font_path = os.path.join(pack_font_path / "roboto mono" / "RobotoMono-Regular.ttf")
default_emoji_font_path = os.path.join(pack_font_path / "emoji" / "NotoEmoji-Regular.ttf")

default_font = pygame.font.Font(default_font_path, 26)
emoji_font = pygame.font.Font(default_emoji_font_path, 26)


class Font:
    def __init__(self, font_path: str = default_font_path, font_size: int = 26):
        self.font = pygame.font.Font(font_path, font_size)

    def __getattr__(self, attr):
        return getattr(self.font, attr)



from setuptools import setup, find_packages

setup(
    name="easypygamewidgets",
    version="26.0",
    packages=find_packages(),
    install_requires=[
        "pygame",
        "requests"
    ],
    include_package_data=True,
    package_data={
        "easypygamewidgets": [
            "assets/fonts/roboto mono/RobotoMono-Regular.ttf",
            "assets/fonts/roboto mono/OFL.txt",
            "assets/fonts/emoji/NotoEmoji-Regular.ttf",
            "assets/fonts/emoji/OFL.txt"
        ]
    },
    author="PizzaPost",
    description="Create GUIs for pygame.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/PizzaPost/pywidgets ",
)
